require 'dmao/api/base'

require 'dmao/api/errors/institution_not_found'
require 'dmao/api/errors/invalid_system_uuid'
require 'dmao/api/errors/invalid_response_length'

module DMAO
  module API

    class Entity < Base

      ENDPOINT = ''
      NOT_FOUND_ERROR = nil
      INVALID_ID_ERROR = nil
      INVALID_ENTITY_CLASS = nil
      INVALID_ENTITY_ERROR_MESSAGE = ""
      VALID_ATTRIBUTES = []

      def self.find_by_system_uuid system_uuid

        validate_system_uuid system_uuid

        response = self.api["#{self::ENDPOINT}?system_uuid=#{system_uuid}"].get

        response_data = JSON.parse(response)["data"]

        raise self::NOT_FOUND_ERROR.new if response_data.length == 0
        raise DMAO::API::Errors::InvalidResponseLength.new("Expected 1 element in response there were #{response_data.length}") if response_data.length != 1

        data = response_data[0]

        instance_from_api_data data

      end

      def self.all

        begin
          response = self.api[self::ENDPOINT].get
        rescue RestClient::NotFound
          raise DMAO::API::Errors::InstitutionNotFound.new
        end

        entities = []

        response_data = JSON.parse(response)["data"]

        return entities if response_data.length == 0

        response_data.each do |data|

          entities.push instance_from_api_data(data)

        end

        entities

      end

      def self.get id

        validate_id id

        begin
          response = self.api["#{self::ENDPOINT}/#{id}"].get
        rescue RestClient::NotFound
          raise self::NOT_FOUND_ERROR.new
        end

        instance_from_response response

      end

      def self.create attributes

        validate_attributes attributes

        begin
          response = self.api[self::ENDPOINT].post attributes.to_json
        rescue RestClient::NotFound
          raise DMAO::API::Errors::InstitutionNotFound.new
        rescue RestClient::UnprocessableEntity => e
          handle_unprocessable_entity e
        end

        instance_from_response response

      end

      def self.update id, attributes

        validate_id id

        validate_attributes attributes

        begin
          response = self.api["#{self::ENDPOINT}/#{id}"].patch attributes.to_json
        rescue RestClient::NotFound
          raise self::NOT_FOUND_ERROR.new
        rescue RestClient::UnprocessableEntity => e
          handle_unprocessable_entity e
        end

        instance_from_response response

      end

      def self.delete id

        validate_id id

        begin
          self.api["#{self::ENDPOINT}/#{id}"].delete
        rescue RestClient::NotFound
          raise self::NOT_FOUND_ERROR.new
        rescue RestClient::UnprocessableEntity => e
          handle_unprocessable_entity e
        end

        true

      end

      def self.instance_from_response response_body

        data = JSON.parse(response_body)["data"]

        instance_from_api_data data

      end

      def self.handle_unprocessable_entity error_response

        errors = JSON.parse(error_response.response.body)["errors"]

        raise self::INVALID_ENTITY_CLASS.new(self::INVALID_ENTITY_ERROR_MESSAGE, errors)

      end

      def self.validate_id id
        if id.nil? || id.to_s.empty?
          raise self::INVALID_ID_ERROR.new
        end
      end

      def self.validate_system_uuid system_uuid
        if system_uuid.nil? || system_uuid.to_s.empty?
          raise DMAO::API::Errors::InvalidSystemUUID.new
        end
      end

      def self.validate_attributes attributes
        attributes.keep_if { |k, _v| self::VALID_ATTRIBUTES.include? k }
      end

      def self.instance_from_api_data _data
        raise "Should be overridden in extending class"
      end

      def self.set_id_attribute_if_not_nil attribute_value
      attribute_value.nil? ? nil : attribute_value["id"]
      end

      def self.parse_error_response error_response
        JSON.parse(error_response.response.body)["errors"]
      end

      def self.raise_error_if_key error_class, errors, key
        raise error_class.new if errors.keys.include? key
      end

    end

  end
end